#
# Author: Michael Fürst
# Version: 1.0
#

require 'socket'
require 'openssl'
require 'json'

class BotConnection
  def initialize(socket, network)
    @socket = socket
    @network = network
  end

  def on_message_handler_loop on_message_handler
    begin
    until closed? do
      on_message_handler.call read
    end
    rescue => e
      puts e.backtrace
    end
  end

  def send msg
    @network.send msg, @socket unless closed?
  end

  def read
    @network.read @socket unless closed?
  end

  def close
    @network.close @socket unless closed?
  end

  def closed?
    @socket.closed?
  end
end

# A worker. Dial home to receive ask for tasks.
class BotNetwork

  attr_reader :socket

  # Initialize the worker.
  def initialize
  end

  def connect(host, port, cert_file = File.open('cert.pem'))
    socket = TCPSocket.open host, port
    socket.setsockopt Socket::IPPROTO_TCP, Socket::TCP_NODELAY, 1
    expectedCert = OpenSSL::X509::Certificate.new(cert_file)
    ssl = OpenSSL::SSL::SSLSocket.new(socket)
    ssl.sync_close = true
    ssl.connect
    if ssl.peer_cert.to_s != expectedCert.to_s
      stderrr.puts 'Unexpected certificate'
      return nil
    end
    @socket = ssl
  end

  def connect_no_ssl(host, port)
    socket = TCPSocket.open host, port
    socket.setsockopt Socket::IPPROTO_TCP, Socket::TCP_NODELAY, 1
    @socket = socket
  end

  def listen(on_connect_handler, port, cert_pem = File.open('cert.pem'), private_pem = File.open('private.pem'))
    server = TCPServer.open(port)
    sslContext = OpenSSL::SSL::SSLContext.new
    sslContext.cert = OpenSSL::X509::Certificate.new(cert_pem)
    sslContext.key = OpenSSL::PKey::RSA.new(private_pem)
    sslServer = OpenSSL::SSL::SSLServer.new(server, sslContext)
    thread = Thread.start do
      puts '[' + Time.new.strftime('%Y-%m-%d %H:%M:%S') + '] ' + "Listening secured on port #{port}"
      accept_clients on_connect_handler, sslServer
    end
    thread
  end

  def listen_no_ssl(on_connect_handler, port)
    server = TCPServer.open(port)
    thread = Thread.start do
      puts '[' + Time.new.strftime('%Y-%m-%d %H:%M:%S') + '] ' + "Listening unsecured on port #{port}"
      accept_clients on_connect_handler, server
    end
    thread
  end

  def accept_clients(on_connect_handler, server)
    loop do
      Thread.start(server.accept) do |socket|
        puts '[' + Time.new.strftime('%Y-%m-%d %H:%M:%S') + '] ' + 'Client connected.!'
        begin
          on_connect_handler.call(BotConnection.new socket, self)
        rescue => e
          puts e.backtrace
          unless socket.closed?
            disconnect socket
          end
        end
        puts '[' + Time.new.strftime('%Y-%m-%d %H:%M:%S') + '] ' + 'Client disconnected!'
      end
    end
  end

  def send(message, socket = @socket)
    socket.puts(JSON.generate(message)) unless socket.closed?
  end

  def read(socket = @socket)
    res = socket.gets
    res = res.chomp if res
    res = JSON.parse(res) if res
    res
  end

  def close(socket = @socket)
    socket.close unless socket.closed?
  end
end